from os.path import exists
import pathlib
import unittest

import numpy as np
from model_server.base.accessors import InMemoryDataAccessor
from model_server.base.session import session

class TestGetSessionObject(unittest.TestCase):
    def setUp(self) -> None:
        session.restart()
        self.sesh = session

    def test_session_logfile_is_valid(self):
        self.assertTrue(exists(self.sesh.logfile), 'Session did not create a log file in the correct place')

    def test_changing_session_root_creates_new_directory(self):
        from model_server.conf.defaults import root

        old_paths = self.sesh.get_paths()
        newroot = root / 'subdir'
        self.sesh.restart(root=newroot)
        new_paths = self.sesh.get_paths()
        for k in old_paths.keys():
            self.assertTrue(new_paths[k].__str__().startswith(newroot.__str__()))

    def test_change_session_subdirectory(self):
        old_paths = self.sesh.get_paths()
        self.sesh.set_data_directory('outbound_images', old_paths['inbound_images'])
        self.assertEqual(self.sesh.paths['outbound_images'], self.sesh.paths['inbound_images'])

    def test_restarting_session_creates_new_logfile(self):
        logfile1 = self.sesh.logfile
        self.assertTrue(logfile1.exists())
        self.sesh.restart()
        logfile2 = self.sesh.logfile
        self.assertTrue(logfile2.exists())
        self.assertNotEqual(logfile1, logfile2, 'Restarting session does not generate new logfile')

    def test_reimporting_session_uses_same_logfile(self):
        session1 = self.sesh
        logfile1 = session1.logfile
        self.assertTrue(logfile1.exists())

        from model_server.base.session import session as session2
        self.assertEqual(session1, session2)
        logfile2 = session2.logfile
        self.assertTrue(logfile2.exists())
        self.assertEqual(logfile1, logfile2, 'Reimporting session incorrectly creates new logfile')

    def test_log_warning(self):
        msg = 'A test warning'
        self.sesh.log_info(msg)
        with open(self.sesh.logfile, 'r') as fh:
            log = fh.read()
        self.assertTrue(msg in log)

    def test_get_logs(self):
        self.sesh.log_info('Info example 1')
        self.sesh.log_warning('Example warning')
        self.sesh.log_info('Info example 2')
        logs = self.sesh.get_log_data()
        self.assertEqual(len(logs), 4)
        self.assertEqual(logs[1]['level'], 'WARNING')
        self.assertEqual(logs[-1]['message'], 'Initialized session')

    def test_change_output_path(self):
        pa = self.sesh.get_paths()['inbound_images']
        self.assertIsInstance(pa, pathlib.Path)
        self.sesh.set_data_directory('outbound_images', pa.__str__())
        self.assertEqual(self.sesh.paths['inbound_images'], self.sesh.paths['outbound_images'])
        self.assertIsInstance(self.sesh.paths['outbound_images'], pathlib.Path)

    def test_make_table(self):
        import pandas as pd
        data = [{'modulo': i % 2, 'times one hundred': i * 100} for i in range(0, 8)]
        self.sesh.write_to_table(
            'test_numbers', {'X': 0, 'Y': 0}, pd.DataFrame(data[0:4])
        )
        self.assertTrue(self.sesh.tables['test_numbers'].path.exists())
        self.sesh.write_to_table(
            'test_numbers', {'X': 1, 'Y': 1}, pd.DataFrame(data[4:8])
        )

        dfv = pd.read_csv(self.sesh.tables['test_numbers'].path)
        self.assertEqual(len(dfv), len(data))
        self.assertEqual(dfv.columns[0], 'X')
        self.assertEqual(dfv.columns[1], 'Y')


    def test_add_and_remove_accessor(self):
        acc = InMemoryDataAccessor(
            np.random.randint(
                0,
                2 ** 8,
                size=(512, 256, 3, 7),
                dtype='uint8'
            )
        )
        shd = acc.shape_dict

        # add accessor to session registry
        acc_id = session.add_accessor(acc)
        self.assertEqual(session.get_accessor_info(acc_id)['shape_dict'], shd)
        self.assertTrue(session.get_accessor_info(acc_id)['loaded'])

        # remove accessor from session registry
        session.del_accessor(acc_id)
        self.assertEqual(session.get_accessor_info(acc_id)['shape_dict'], shd)
        self.assertFalse(session.get_accessor_info(acc_id)['loaded'])

    def test_add_and_use_accessor(self):
        acc = InMemoryDataAccessor(
            np.random.randint(
                0,
                2 ** 8,
                size=(512, 256, 3, 7),
                dtype='uint8'
            )
        )
        shd = acc.shape_dict

        # add accessor to session registry
        acc_id = session.add_accessor(acc)
        self.assertEqual(session.get_accessor_info(acc_id)['shape_dict'], shd)
        self.assertTrue(session.get_accessor_info(acc_id)['loaded'])

        # get accessor from session registry without popping
        acc_get = session.get_accessor(acc_id, pop=False)
        self.assertIsInstance(acc_get, InMemoryDataAccessor)
        self.assertEqual(acc_get.shape_dict, shd)
        self.assertTrue(session.get_accessor_info(acc_id)['loaded'])

        # get accessor from session registry with popping
        acc_get = session.get_accessor(acc_id)
        self.assertIsInstance(acc_get, InMemoryDataAccessor)
        self.assertEqual(acc_get.shape_dict, shd)
        self.assertFalse(session.get_accessor_info(acc_id)['loaded'])
